#!/usr/bin/perl
# ------------------------------------------------------------------------------
#
# File: wrapper.pl
#
# Written: M Gerrard, 2003 approx
#
# This program is a wrapper for the compiler.
# It extracts command line parameters that could be applicable to QAC or QAC++
# the command line for analysis
#
# ------------------------------------------------------------------------------
#  version             change
#  2.6                 CR 12840 Wrapper overhauled. PROJECT_CREATOR_MODE added.
#  2.5                 CR 12501 Visa Eruope. Output to be sent to different dir,
#                      Configurable defines on compiler options, configurable
#                      include option (-i doesn't mean include on Sun compiler)
#                      SLEEP_AFTER_ANALYSIS now supports sleeps of less than one
#                      second
#  2.4                 CR 12032 On windows and cygwin, Include paths starting
#                      with a / (ie UNIX paths) are converted to Windows paths.
#                      This also resolves symlinked directories.
#                      CR 12033 Empty -I and -D options are ignored and do not
#                      consume the next argument
#                      CR 12034 SLEEP_AFTER_ANALYSIS option add to configuration
#                      options to pause wrapper after primary analysis 
#                      CR 12036 Quoted strings in parameters (typically -D) are
#                      now handled correctly - this is based on the OS. For
#                      Windows Perl only (ie wrapper.exe) the quotes are
#                      replaced. An additional bug whereby they were replaced
#                      twice has been removed.
#                      Extra debug information regarding filename removes.
#  2.3                 Generate output / filelist regardless of analysis failure
#                      Remove warning of non-escaped "@" in source.
# ------------------------------------------------------------------------------

use strict;

our $version = "2.6";           # Version number

our $mode;                      # Controls operation of this wrapper.pl
                                # Value is "PROJECT_CREATOR_MODE" or
                                # "MAKEFILE_INTEGRATION_MODE"

our $wqp_list_file_path;        # Path to location of wqp_list.  Only relevant
                                # in PROJECT_CREATOR_MODE mode.
                                # NOTE: FILE $wqp_list_file_path MUST BE
                                # DELETED BY THE PROGRAM WHICH INVOKES make
                                # PRIOR TO INVOKING make.

our $compilerExecutable;        # The path to the compiler used by the wrapper.

our $compilerSettingsFile;      # This is usally the compiler personality file

our $integrationType;           # integrationType describes the mode of operation
                                # of the wrapper.  It can be either
                                # nothing "", STDERR, TEXT, HTML, STDERR_AND_TEXT,
                                # STDERR_AND_HTML, or STDERR_TEXT_AND_HTML

our $useLogging;                # Set to 1 if you want to log the analysis/compilation
                                # otherwise it is sent to stderr

our $logFile;                   # This is the output file used when $useLogging is 1.

our $debug = 0;                 # This gives extra debug information if set

our $stopOnFail;                # stopOnFail determines whether a failed analysis 
                                # should return a non-zero exit code

our $analyser;                  # This is either QAC or QACPP
our $analyserBase;              # This is the path where QAC/QAC++ is installed
our $analyserFlags;             # These are the flags for the analyser
our $errdspFlags;               # These are the flags for errdsp
our $postAnalysis;              # This is the post-analysis program
our $sleepAfterAnalysis;        # Optional pause after each analysis
our $fileList;                  # Names the LOF file which is optionally created
                                # NOTE: FILE $fileList MUST BE
                                # DELETED BY THE PROGRAM WHICH INVOKES make
                                # PRIOR TO INVOKING make.

our $altOutputDir;              # This is the name of the alternate output dir for
                                # analysis op CR 12501

our %defineOnOption;            # Table of defines for analysis if compiler option
                                # set CR 12501

our $includeFlag;               # Set to allowable include flags CR 12501

# derived settings
#

our $analyserExecutable;        # This is the path to the analyser
our $errdspExecutable;          # This is the path to errdsp

our $outputDirectory;           # This is set to the same as the source file

our @compilerCommandLine;

our @qa_options;                # These are the options that are produced
                                # from parseCommandLine

our $sourceFile;                # These are the source files that are found
                                # from parseCommandLine

our $LOGFILE;

our $unix;
our $windows_or_cygwin;
our $pure_windows;
our $cygwin;
our $relativeDir;
our $pwdAtStart;

# ------------------------------------------------------------------------------
#
# main
#
# ------------------------------------------------------------------------------

  $| = 1; # flush stdout after each print

  my $analysisResult = 0;
  
  checkUnixOrWindows();
  
  $pwdAtStart = PWD();
  $relativeDir = $pwdAtStart;

  if ( $windows_or_cygwin )
  {
      if ( WINDOWS_ONLY_WHICH_PROGRAM( "cygpath" ) eq "" )
      {
          DIE( "wrapper unable to run cygpath.\nEnsure cygpath is on PATH" );
      }
  }

  # read the settings
  
  my $parse_successful = parseCommandLine();

  if ( $parse_successful )
  {
      setupEnvironmentVariables();
  
      $analysisResult = runAnalysis();
  
      if ( ( $analysisResult == 0 ) &&
           ( $postAnalysis ne "" )
         )
      {
          runPostAnalysis();
      } 
      elsif ( $postAnalysis ne "" )
      {
          # Analysis failed, log skipping of post analysis
          #
          logMessage( "*** Analysis Failed: No post-analysis performed ***\n" );
      }
      
      # Generate output so that hard errors can be seen
      #
      generateOutput();  
  }
  else
  {
      logMessage( "*** No analysis performed ***\n" );
  }
  
  my $exitValue = 0;

  if ( $stopOnFail && $analysisResult )
  {
      logMessage( "*** ABORTING COMPILATION AFTER ANALYSIS FAILURE (STOP_ON_FAIL=1) ***\n" );
      $exitValue = $analysisResult;
  }
  else
  {
      if ( $analysisResult )
      {
          logMessage( "CONTINUING COMPILATION AFTER ANALYSIS FAILURE (STOP_ON_FAIL=0)\n" );
      }
      $exitValue = runCompilation();
  }
  
  # Close the log file
  #
  closeLogFile();
  
  if ( $exitValue )
  {
      exit 255;
  }
  else
  {
      exit 0;
  }
    
# ------------------------------------------------------------------------------

sub checkUnixOrWindows
{
    $cygwin = ( $^O =~ /cygwin/o );
    $pure_windows = ( $^O =~ /MSWin/o );

    if ( $pure_windows || $cygwin )
    {
        $unix = 0;
        $windows_or_cygwin = 1;
    }
    else
    {
        $unix = 1;
        $windows_or_cygwin = 0;
    }
}

# ------------------------------------------------------------------------------

sub usage
{
    logMessage( "Command Line Wrapper for QAC/QAC++ version $version\n" );
    logMessage( "Copyright Programming Research Ltd 2008\n" );
    logMessage( "\n" );
    logMessage( "Usage:\n" );
    logMessage( "wrapper [-wcf<FILE>|-wcf <FILE>]+ " .
                "[-setOPTION=VALUE|-set OPTION=VALUE]+ " .
                "<compiler> <compiler command line> <file>\n"
              );
    logMessage( "\n" );
    logMessage( "For detailed information, consult PDF manual\n" );
    logMessage( "For support, please contact support\@programmingresearch.com\n" ); 
    exit;
}

# ------------------------------------------------------------------------------

sub setDefaultSettings
{
    $mode                   = "MAKEFILE_INTEGRATION_MODE"; # default value
    $wqp_list_file_path     = ""; # must be set up
    $compilerSettingsFile   = ""; # default nothing (must be set)
    $integrationType        = ""; # default value
    $useLogging             = 0;  # default value
    $logFile                = ""; # default value
    $stopOnFail             = 0;  # default value
    $analyser               = ""; # must be set up
    $analyserBase           = ""; # must be set up
    $analyserFlags          = ""; # default value
    $errdspFlags            = ""; # default value
    $postAnalysis           = ""; # default value
    $sleepAfterAnalysis     = 0;  # default no sleep
    $fileList               = ""; # default value
    $altOutputDir           = ""; # default is for op to go where make puts it's op
    %defineOnOption         = (); # no extra defines by default
    $includeFlag            = "[I|i]"; # default is upper or lower case i

    # derived

    $analyserExecutable     = ""; # will be set up
    $errdspExecutable       = ""; # will be set up
}

# ------------------------------------------------------------------------------
#
# Configuration options can contain environment variables, which must be 
# expanded.  
# Also, where the option ( ~ OUTPUT_DIR ) is a directory or file, this can be
# in the form of a relative path.  These paths are relative to the current
# directory, or if the option is set in a configuration file it is relative
# to the location of the configuration file.
#
sub expandOption
{
    my $baseOption = shift;

    # expand $(X) type environment variables
    #
    my $t = EXPAND_ENV_VARS( $baseOption );

    # expand relative paths
    #
    $t = EXPAND_RELATIVE_PATHS( $t );

    return $t;
}

# ------------------------------------------------------------------------------

sub EXPAND_ENV_VARS
{
    # Usage $expanded_var = EXPAND_ENV_VARS( $var );
    #
    my $var = shift;

    # expand $(X) type environment variables
    #

    while ( $var =~ /^(.*?)\$\((.*?)\)(.*)$/o )
    {
        my $env_var = $2;

        if ( ! exists( $ENV{ $env_var } ) )
        {
            logMessage( "Environment variable $env_var is not set\n" );
            $var = $1 . $3;
        }
        else
        {
            $var = $1 . $ENV{ $env_var } . $3;
        }
    }

    return $var;
}

# ------------------------------------------------------------------------------

sub EXPAND_RELATIVE_PATHS
{
    # Usage: $expanded_path = EXPAND_RELATIVE_PATHS( $path );
    #
    my $path = shift;

    if ( $path =~ /^\./o )
    {
        return APPLY_FINAL_SEP( $relativeDir ) . $path;
    }
    else
    {
        return $path;
    }
}

# ------------------------------------------------------------------------------
#
# This function sets a wrapper option based on a single setting
#
sub setupOption
{
    my $line = shift;

    if ( $line =~ /^MODE=(.*)$/o )
    {
        $mode = expandOption( $1 );
    }
    elsif ( $line =~ /^WQP_LIST_FILE_PATH=(.*)$/o )
    {
        $wqp_list_file_path = expandOption( $1 );
    }
    elsif ( $line =~ /^INCLUDE_OPTION=(.*)$/o )  # CR 12501
    {
        $includeFlag = expandOption( $1 );
    }
    elsif ( $line =~ /^DEFINE_ON_OPTION=(.*)$/o )  # CR 12501
    {
        %defineOnOption = PARSE_DEFINE_OPTIONS( $1 );
    }
    elsif ( $line =~ /^OUTPUT_DIR=(.*)$/o )  # CR 12501
    {
        $altOutputDir = EXPAND_ENV_VARS( $1 );
    }
    elsif ( $line =~ /^SLEEP_AFTER_ANALYSIS=(.*)$/o )
    {
        $sleepAfterAnalysis = expandOption( $1 ); # overrides previous settings
    }
    elsif ( $line =~ /^POST_ANALYSIS=(.*)$/o )
    {
        $postAnalysis = expandOption( $1 ); # overrides previous settings
    }
    elsif ( $line =~ /^FILELIST=(.*)$/o )
    {
        $fileList = expandOption( $1 ); # overrides previous settings
    }
    elsif ( $line =~ /^COMPILER_SETTINGS_FILE=(.*)$/o )
    {
        $compilerSettingsFile = expandOption( $1 ); # overrides previous settings
    }
    elsif ( $line =~ /^INTEGRATION_TYPE=(.*)$/o )
    {
        $integrationType = expandOption( $1 );
    }
    elsif ( $line =~ /^LOGFILE=(.*)$/o )
    {
        # once set, this cannot be unset by subsequent 
        # wrapper config files
        $useLogging = 1;
        $logFile = expandOption( $1 );
    }
    elsif ( $line =~ /^ANALYSER=(.*)$/o )
    {
        # setting ANALYSER overrides previous settings
        #
        if ( uc( $1 ) eq "QAC" )
        {
            $analyser = "QAC";
        }
        else
        {
            $analyser = "QACPP";
        }
    }
    elsif ( $line =~ /^ANALYSER_BASE=(.*)$/o )
    {
        # setting ANALYSER_BASE overrides previous settings
        #
        $analyserBase = expandOption( $1 );
    }
    elsif ( $line =~ /^ANALYSER_FLAGS=(.*)$/o )
    {
        # As can have: ANALYSER_FLAGS=$ANALYSER_FLAGS flag flag ...
        # we need to replace $ANALYSER_FLAGS with $analyserFlags
        #
        my $newAnalyserFlags = expandOption( $1 );
        $newAnalyserFlags =~ s/\$ANALYSER_FLAGS/$analyserFlags/g;
        $analyserFlags = $newAnalyserFlags;
    }
    elsif ( $line =~ /^ERRDSP_FLAGS=(.*)$/o )
    {
        # As can have: ERRDSP_FLAGS=$ERRDSP_FLAGS flag flag ...
        # we need to replace $ERRDSP_FLAGS with $errdspFlags
        #
        my $newErrdspFlags = expandOption( $1 );
        $newErrdspFlags =~ s/\$ERRDSP_FLAGS/$errdspFlags/g;
        $errdspFlags = $newErrdspFlags;
    }
    elsif ( $line =~ /^DEBUG=(.*)$/o )
    {
        # setting DEBUG overrides previous settings
        #
        if ( expandOption( $1 ) eq "1" )
        {
            $debug = 1;
        }
        else
        {
            $debug = 0;
        }
    }
    elsif ( $line =~ /^STOP_ON_FAIL=(.*)$/o )
    {
        # setting STOP_ON_FAIL overrides previous settings
        #
        if ( expandOption( $1 ) eq "1" )
        {
            $stopOnFail = 1;
        }
        else
        {
            $stopOnFail = 0;
        }
    }
}

# ------------------------------------------------------------------------------
#
# This function reads a settings file
#
sub readSettingsFile
{
    my $settingsFile = shift;

    $settingsFile = GET_CORRECT_PATH( $settingsFile );

    open ( SETTINGS, "<$settingsFile" ) || DIE( "Cannot open $settingsFile\n$!" );
    
    $relativeDir = SHORT_PATH_NAME( DIR_NAME( $settingsFile ) );

    while ( my $line = <SETTINGS> )
    {
        $line = CHOMP( $line );
        setupOption( $line );
    }

    close SETTINGS;

    $relativeDir = $pwdAtStart;
}

# ------------------------------------------------------------------------------

sub validateSettings
{
    openLogFile();

    logMessage( "Command Line Wrapper for QAC/QAC++ version $version\n" );
    logMessage( "Copyright Programming Research Ltd 2008\n" );

    if ( $analyser eq "QACPP" )
    {
        # windows it is qacpp, unix it is qac++.        

        if ( $windows_or_cygwin )
        {
            $analyserExecutable = $analyserBase . "/bin/qacpp.exe";
        }
        else
        {
            $analyserExecutable = $analyserBase . "/bin/qac++";
        }
    }
    else 
    {
        if ( $windows_or_cygwin )
        {
            $analyserExecutable = $analyserBase . "/bin/qac.exe";
        }
        else
        {
            $analyserExecutable = $analyserBase . "/bin/qac";
        }
    }

    if ( $windows_or_cygwin )
    {
        $errdspExecutable = $analyserBase . "/bin/errdsp.exe";
    }
    else
    {
        $errdspExecutable = $analyserBase . "/bin/errdsp";
    }

    $analyserExecutable   = SHORT_PATH_NAME( $analyserExecutable );
    $errdspExecutable     = SHORT_PATH_NAME( $errdspExecutable );
    $compilerSettingsFile = SHORT_PATH_NAME( $compilerSettingsFile );

    if ( $postAnalysis ne "" )
    {
       $postAnalysis = SHORT_PATH_NAME( $postAnalysis );
    }
    
    my $returnValue = 1;

    if ( ( $mode eq "MAKEFILE_INTEGRATION_MODE" ) &&
         ( $compilerSettingsFile eq "" ) 
       )
    {
        logMessage( "*** Required setting for COMPILER_SETTINGS_FILE is not set ***\n" );
        $returnValue = 0;
    }
    else
    {
        debugMessage( "COMPILER_SETTINGS_FILE=$compilerSettingsFile\n" );
    }

    if ( $analyser eq "" )
    {
        logMessage( "*** Required setting for ANALYSER is not set ***\n" );
        $returnValue = 0;
    }
    else
    {
        debugMessage( "ANALYSER=$analyser\n" );
    }

    if ( $analyserBase eq "" )
    {
        logMessage( "*** Required setting for ANALYSER_BASE is not set ***\n" );
        $returnValue = 0;
    }
    else
    {
        debugMessage( "ANALYSER_BASE=$analyserBase\n" );
    }

    if ( ( $mode ne "MAKEFILE_INTEGRATION_MODE" ) &&
         ( $mode ne "PROJECT_CREATOR_MODE" )
       )
    {
        logMessage( "*** MODE must be MAKEFILE_INTEGRATION_MODE or " .
                    "PROJECT_CREATOR_MODE ***\n" );
        $returnValue = 0;
    }

    if ( ( $mode eq "PROJECT_CREATOR_MODE" ) &&
         ( $wqp_list_file_path eq "" )
       )
    {
        logMessage( "*** WQP_LIST_FILE_PATH must be specified when in " .
                    "PROJECT_CREATOR_MODE ***\n" );
        $returnValue = 0;
    }


    if ( $returnValue )
    {
        debugMessage( "INTEGRATION_TYPE=$integrationType\n" );
        debugMessage( "DEBUG=$debug\n" );
        
        if ( $useLogging )
        {
            debugMessage( "LOGFILE=$logFile\n" );
        }

        debugMessage( "ANALYSER_FLAGS=$analyserFlags\n" );
        debugMessage( "ERRDSP_FLAGS=$errdspFlags\n" );
        debugMessage( "POST_ANALYSIS=$postAnalysis\n" );
        debugMessage( "Analyser is $analyserExecutable\n" );
        debugMessage( "Errdsp is $errdspExecutable\n" );
    }

    return $returnValue;
}

# ------------------------------------------------------------------------------
#
# setup environment variables
#
sub setupEnvironmentVariables
{ 
    # set up the bin directory QACBIN or QACPPBIN
    #
    $ENV{ $analyser . "BIN" } = SHORT_PATH_NAME( $analyserBase . "/bin" );

    # set up  QACOUTPATH or QACPPOUTPATH
    #
    $ENV{ $analyser . "OUTPATH" } = PWD();
    
    # set up QACHELPFILES or QACPPHELPFILES
    #
    $ENV{ $analyser . "HELPFILES" } = SHORT_PATH_NAME( $analyserBase . "/help" );
}

# ------------------------------------------------------------------------------
#
# openLogFile
#
sub openLogFile
{
    if ( $useLogging )
    {
        open LOGFILE, ">>$logFile";
    }
}

# ------------------------------------------------------------------------------
#
# logMessage
#
sub logMessage
{
    my $message = shift;

    if ( $useLogging )
    {
        print LOGFILE ">> " . $message;
    }
    else
    {
        print STDERR ">> " . $message;
    }
}

# ------------------------------------------------------------------------------
#
# debugMessage
#
sub debugMessage
{
    my $message = shift;

    if ( $debug )
    {
        logMessage( $message );
    }
}

# ------------------------------------------------------------------------------
#
# closeLogFile
#
sub closeLogFile
{
    if ( $useLogging )
    {
        close LOGFILE;
    }
}

# ------------------------------------------------------------------------------
#
# parseCommandLine parses the command line and sets up @qa_options
# It will return 1 if the command line was deemed a compilation line
# i.e some analysis was required.
#
sub parseCommandLine
{
    my $nextArgumentIsDefine            = 0; # set to 1 if the argument is a sole -D|-d
    my $nextArgumentIsInclude           = 0; # set to 1 if the argument is a sole -I|-i
    my $nextArgumentIsObjectFile        = 0; # set to 1 if the argument is a sole -o
    my $nextArgumentIsOptionsFile       = 0; # set to 1 if the argument is a sole -f
    my $nextArgumentIsWrapperConfigFile = 0; # set to 1 if the argument is a sole -wcf
    my $nextArgumentIsSetting           = 0; # set to 1 if the argument is a sole -set
    my @qa_includes = ();    # the includes
    my @qa_defines  = ();    # the defines
    my @qa_files    = ();    # source files
    my @CARGS       = ();    # the compiler arguments

    # Set up flagRegex to represent a regular expression
    # for options on windows/unix
    #
    my $flagRegex;
    if ( $windows_or_cygwin )
    {
        $flagRegex = "[-\/]";
    }
    else
    {
        $flagRegex = "-";
    }

    # Set the default settings prior to reading the wrapper
    # config files
    #
    setDefaultSettings();

    # First, parse the command line in order to pick
    # up the -wcf & -set options and the compiler
    #
    foreach my $arg ( @ARGV )
    {
        if ( $nextArgumentIsWrapperConfigFile )
        {
            # Read the settings from the wrapper config file
            #
            readSettingsFile( $arg );
            $nextArgumentIsWrapperConfigFile = 0;
        }
        elsif ( $nextArgumentIsSetting )
        {
            # Set the setting as specified 
            #
            setupOption( $arg );
            $nextArgumentIsSetting = 0;
        }
        elsif ( $arg =~ /^${flagRegex}wcf$/ )
        {
            $nextArgumentIsWrapperConfigFile = 1;
        }
        elsif ( $arg =~ /^${flagRegex}wcf(.+)$/ )
        {
            readSettingsFile( $1 );
        }
        elsif ( $arg =~ /^${flagRegex}set$/ )
        {
            $nextArgumentIsSetting = 1;
        }
        elsif ( $arg =~ /^${flagRegex}set(.+)$/ )
        {
            setupOption( $1 );
        }
        elsif ( $arg =~ /^${flagRegex}v$/ ||
                $arg =~ /^${flagRegex}version$/ ||
                $arg =~ /^${flagRegex}\?$/ )
        {
            usage();
        }
        else
        {
            push @CARGS, $arg;
        }
    }

    # Now we chop off the first argument (the compiler executable)
    #
    ( $compilerExecutable, @CARGS ) = @CARGS;

    # Set the output directory to the current directory
    # (this is the default behaviour for most compilers).
    # If we see a "-o" option we will use that directory instead.
    #
    $outputDirectory = PWD();

    # Construct the compiler command line (this must be done b4 the call
    # to validateSettings)
    # 
    @compilerCommandLine = @CARGS;

    # validate that the settings are ok
    #
    if ( validateSettings() == 0 )
    {
        logMessage( "*** Incomplete settings ***\n" );
        return 0;
    }

    my $index = 0;

    CARG:
    while ( $index <= $#CARGS )
    {
        my $arg = $CARGS[ $index ];

        my $next_arg;
        if ( $index < $#CARGS )
        {
            $next_arg = $CARGS[ $index + 1 ];
        }
        else
        {
            $next_arg = "";
        }

        if ( $nextArgumentIsDefine ) 
        {
            push @qa_defines, $arg;

            $nextArgumentIsDefine = 0;
        }
        elsif ( $nextArgumentIsInclude ) 
        {
            # CR12032 -I<space>path convert UNIX paths to windows paths on cygwin
            #
            push @qa_includes, GET_CORRECT_PATH( $arg );

            $nextArgumentIsInclude = 0;
        }
        elsif ( $nextArgumentIsObjectFile )
        {
            my $output_directory_supplied = DIR_NAME( GET_FILE_PATH_ABSOLUTE( $arg ) );

            # CR 12501 only change op dir if option selected 

            if ( $altOutputDir ne "" )
            {
                $outputDirectory = MODIFY_OUT_PATH( $output_directory_supplied, $altOutputDir );
            }
            else
            {
                $outputDirectory = $output_directory_supplied;
            }

            $outputDirectory = GET_CORRECT_PATH( $outputDirectory );

            MK_PATH( $outputDirectory );

            $nextArgumentIsObjectFile = 0;
        }
        elsif ( $nextArgumentIsOptionsFile )
        {
            @CARGS = LOAD_OPTIONS_FILE( $arg, $index, @CARGS );
            
            $nextArgumentIsOptionsFile = 0;
        }            

        elsif ( $arg =~ /^\@(.+)$/o )
        {
            # Microsoft CL @command-file (like QAC -via)
            #
            @CARGS = LOAD_OPTIONS_FILE( $1, $index, @CARGS );
        }            
        elsif ( $arg =~ /^${flagRegex}f$/ )
        {
            $nextArgumentIsOptionsFile = 1;
        }
        elsif ( $arg =~ /^${flagRegex}o$/ ) 
        {
            $nextArgumentIsObjectFile = 1;
        }
        elsif ( $windows_or_cygwin && 
                ( $arg =~ /^${flagRegex}Fo$/ )
              )
        {
            $nextArgumentIsObjectFile = 1;
        }
        elsif ( $arg =~ /^${flagRegex}[D|d]$/ ) 
        {
            # CR10233 missing path after include - employ fix for defines as well
            if ( $next_arg =~ /^${flagRegex}/ )
            {
                logMessage( "Empty define directive ignored\n" );
            }
            else  # the next argument is a macro   
            { 
                $nextArgumentIsDefine = 1;
            }
        }
        elsif ( $arg =~ /^${flagRegex}$includeFlag.*$/ ) 
        {
            # CR12501 allow specification of include flags
            #
            if ( $arg !~ /^${flagRegex}$includeFlag$/ )
            {
                my $new_arg = "";
                my $nch = length( $arg );

                while ( substr( $arg, 0, $nch ) !~ /^${flagRegex}$includeFlag$/ )
                {
                    --$nch;
                    $new_arg = substr( $arg, $nch );
                }

                push @qa_includes, GET_CORRECT_PATH( $new_arg );
            }
            else
            {
                # CR10233 missing path after include.
                # Check that argument is not an option.
                # Can't check -I <blank> <file to compile>
                #
                if ( $next_arg =~ /^${flagRegex}/ )
                {
                    logMessage( "Empty include directive ignored\n" );
                }
                else
                {
                    # Next argument is a path
                    #
                    $nextArgumentIsInclude = 1;  # true
                }
            }
        }

        elsif ( $arg =~ /^${flagRegex}[D|d](.+)$/ ) 
        {
            push @qa_defines, $1;
        }
        elsif ( $windows_or_cygwin && ( $arg =~ /^${flagRegex}Fo(.+)$/ ) )
        {
            my $output_directory_supplied = DIR_NAME( GET_FILE_PATH_ABSOLUTE( $1 ) );

            # CR 12501 only change op dir if option selected 

            if ( $altOutputDir ne "" )
            {
                $outputDirectory = MODIFY_OUT_PATH( $output_directory_supplied, $altOutputDir );
            }
            else
            {
                $outputDirectory = $output_directory_supplied;
            }

            $outputDirectory = GET_CORRECT_PATH( $outputDirectory );

            MK_PATH( $outputDirectory );
        }

        elsif ( $arg =~ /^[^-].*\.cpp$/o ) 
        {
            push @qa_files, $arg;
        }
        elsif ( $arg =~ /^[^-].*\.cxx$/o ) 
        {
            push @qa_files, $arg;
        }
        elsif ( $arg =~ /^[^-].*\.[c|C]$/o ) 
        {
            push @qa_files, $arg;
        }
        elsif ( $arg =~ /^[^-].*\.cc$/o ) 
        {
            push @qa_files, $arg;     
        }

        elsif ( $arg =~ /^${flagRegex}MM/ )
        {
            # it is not a real compilation, so we defer the analysis
            #
            return 0;
        }
        elsif ( $windows_or_cygwin && ( $arg =~ /^${flagRegex}MD$/ ) )
        {
            push @qa_defines, "_MT";
            push @qa_defines, "_DLL";
        }
        elsif ( $windows_or_cygwin && ( $arg =~ /^${flagRegex}MDd$/ ) )
        {
            push @qa_defines, "_MT";
            push @qa_defines, "_DLL";
            push @qa_defines, "_DEBUG";
        }
        elsif ( $windows_or_cygwin && ( $arg =~ /^${flagRegex}MT$/ ) )
        {
            push @qa_defines, "_MT";
        }
        elsif ( $windows_or_cygwin && ( $arg =~ /^${flagRegex}MTd$/ ) )
        {   
            push @qa_defines, "_MT";
            push @qa_defines, "_DEBUG";
        }
        elsif ( $windows_or_cygwin && ( $arg =~ /^${flagRegex}MLd$/ ) )
        {   
            push @qa_defines, "_DEBUG";
        }
        elsif ( $windows_or_cygwin && ( $arg =~ /^${flagRegex}LDd$/ ) )
        {   
            push @qa_defines, "_DEBUG";
        }
        elsif ( $windows_or_cygwin && ( $arg =~ /^${flagRegex}GR$/ ) )
        {
            push @qa_defines, "_CPPRTTI";
        }

        # Run through the list of keys in the option list to see 
        # if there is a replacement defined for this $arg option.
        #
        while ( my ( $key, $val ) = each( %defineOnOption ) )
        {
            if ( $arg =~ /${flagRegex}$key$/ )
            {
                push @qa_defines, $val;
            }
        }
    }
    continue
    {
        ++$index;
    }

    if ( scalar( @qa_files ) == 1 )
    {
        $sourceFile = $qa_files[0];
    }
    else
    {
        # Not handled support for more than 1 file or no files
        #
        logMessage( "*** The wrapper supports analysis of only one file at a time ***\n" );
        return 0;
    }

    push @qa_options, ( "-op", $outputDirectory );
    
    # Expand fileList
    #
    if ( $fileList ne "" )
    {
        $fileList = APPLY_FINAL_SEP( $outputDirectory ) . $fileList;
    }

    for my $include ( @qa_includes )
    {
        # make all include paths absolute
        #
        if ( $include ne "" )
        {
            if ( ! IS_FILE_PATH_ABSOLUTE( $include ) )
            {
                $include = SHORT_PATH_NAME( GET_FILE_PATH_ABSOLUTE( $include ) );
            }

            push @qa_options, ( "-I", $include );
        }
    }

    foreach my $define ( @qa_defines )
    {
        push @qa_options, ( "-D", $define );
    }

    # The compiler settings file *must* come after the includes
    # specified on the command line
    #
    if ( $compilerSettingsFile ne "" )
    {
        push @qa_options, ( "-via", $compilerSettingsFile );
    }

    if ( $analyserFlags ne "" )
    {
        push @qa_options, SPLIT_STRING_ALLOWING_QUOTED_TERMS( $analyserFlags );
    }

    return 1;
}


# ------------------------------------------------------------------------------

sub LOAD_OPTIONS_FILE
{
    # Usage: @CARGS = LOAD_OPTIONS_FILE( $option_filename, $index, @CARGS );
    #
    # Called in middle of parsing $index'th element of @CARGS options 
    # to insert options immediately after the $index element of @CARGS.
    # Inserts options found in $option_filename into @CARGS.
    # No options are removed from @CARGS as preceding @CARGS options
    # are no longer of any interest in the calling parseCommandLine loop.
    #
    my $option_filename = shift;
    my $index           = shift;
    my @CARGS           = @_;

    my @opt_lines = LOAD_FILE_AS_ARRAY( $option_filename );

    my @olist = ();
    for my $line ( @opt_lines )
    {
        push @olist, SPLIT_STRING_ALLOWING_QUOTED_TERMS( $line );
    }

    splice ( @CARGS, $index + 1, 0, @olist );

    return @CARGS;
}

# ------------------------------------------------------------------------------
#
# runPostAnalysis - only called if there is post-analysis to run
#
sub runPostAnalysis
{
    # Require here $mode is "MAKEFILE_INTEGRATION_MODE"
    # 
    if ( $mode ne "MAKEFILE_INTEGRATION_MODE" ) 
    {
        return 0;
    }

    my @commandLine = ( $postAnalysis, "-op", $outputDirectory, $sourceFile );

    debugMessage( join( " ", @commandLine ) . "\n" );

    logMessage( "Post-Analysing $sourceFile ...\n" );
    
    my $analysisExitValue = EXECUTE_ARGS_WITH_RESULT( @commandLine );

    if ( $analysisExitValue != 0 ) 
    {
        logMessage( "Post Analysis FAILED\n" );
        logMessage( "*** Post-analysis of $sourceFile failed ($analysisExitValue). ***\n" );
    }
    else 
    {
        logMessage( "Post Analysis OK\n" );
    }

    return $analysisExitValue;
}

# ------------------------------------------------------------------------------
#
# runAnalysis
#
sub runAnalysis
{
    if ( $mode eq "PROJECT_CREATOR_MODE" )
    {
        # Create a file named "$sourceFile.wqp", write the sourceFile
        # and options to it.
        #
        Write_WQP_File( $sourceFile, @qa_options );

        return 0;
    }

    # Assert here $mode is "MAKEFILE_INTEGRATION_MODE"
    # 
    if ( $mode ne "MAKEFILE_INTEGRATION_MODE" ) 
    {
        DIE( "Program Error in runAnalysis()" );
    }

    logMessage( "Analysing $sourceFile ...\n" );

    debugMessage( "$analyserExecutable " . join( " ", @qa_options ) . " $sourceFile\n" );

    my $analysisExitValue = EXECUTE_ARGS_WITH_RESULT( $analyserExecutable, @qa_options, $sourceFile );

    my @errdsp_flags = SPLIT_STRING_ALLOWING_QUOTED_TERMS( $errdspFlags );

    my $temp_file = APPLY_FINAL_SEP( $outputDirectory ) . BASE_NAME( $sourceFile ) . ".tmp";

    if ( $analyser eq "QAC" )
    {
        # QAC is broke!
        # it always returns an exit value of 0 regardless of whether it had
        # any hard errors.
        # the only to work out if we have hard errors is to grep through
        # the output from errdsp!
        #
        my $hardErrorFormat = "%?h%(HARDERROR%)";

        my $exitValue = EXECUTE_ARGS_WITH_RESULT
                           ( $errdspExecutable, $analyser, "-HDR-", "-SU0-8",
                             "-op", $outputDirectory, @errdsp_flags,
                             "-messagesonly+", "-format", $hardErrorFormat,
                             $sourceFile, "-file", $temp_file );

        my @temp_file = LOAD_FILE_AS_ARRAY( $temp_file );

        if ( grep( /HARDERROR/o, @temp_file ) > 0 )
        {
            $analysisExitValue = 1;
        }
    }

    if ( $analysisExitValue != 0 ) 
    {
        logMessage( "Analysis FAILED\n" );

        my $exitValue = EXECUTE_ARGS_WITH_RESULT
                           ( $errdspExecutable, $analyser, "-HDR-", "-SU0-8",
                             "-op", $outputDirectory, @errdsp_flags,
                             "-messagesonly+", $sourceFile, "-file", $temp_file );

        logMessage( "*** Problems analysing $sourceFile. ***\n" );
        logMessage( "*** HARD errors detected indicating that the analyser failed " .
                    "to analyse your code successfully.  ***\n" );
        logMessage( "*** Check your compiler configuration ***\n" );

        my @temp_file = LOAD_FILE_AS_ARRAY( $temp_file );

        for my $line ( @temp_file )
        {
            logMessage( $line );
        }
    }
    else 
    {
        logMessage( "Analysis OK\n" );
    }

    # Optional sleep after analysis completed
    #
    if ( $sleepAfterAnalysis != 0 )
    {
        my $mess;
        if ( $sleepAfterAnalysis == 1 )
        {
            $mess = "Sleep for 1 second ...\n";
        }
        else
        {
            $mess = "Sleep for $sleepAfterAnalysis seconds ...\n";
        }

        debugMessage( $mess );

        sleep $sleepAfterAnalysis;
    }  
 
    return $analysisExitValue;
}

# ------------------------------------------------------------------------------
#
# runCompilation
#
sub runCompilation
{
    # Require here $mode is "MAKEFILE_INTEGRATION_MODE"
    # 
    if ( $mode ne "MAKEFILE_INTEGRATION_MODE" ) 
    {
        return 0;
    }

    # Run the compiler

    logMessage( "Compiling ...\n" );

    # Run the compiler and check to see whether we get a non-zero
    # exit code. If we do, exit and return that code.
    #
    debugMessage( "$compilerExecutable " . join( " ", @compilerCommandLine ) . "\n" );

    my $exitValue = EXECUTE_ARGS_WITH_RESULT( $compilerExecutable, @compilerCommandLine );

    if ( $exitValue != 0 ) 
    {
        logMessage( "Compilation FAILED\n" );
    }
    else
    {
        logMessage( "Compilation OK\n" );
    }

    return $exitValue;
}

# ------------------------------------------------------------------------------
#
# generateOutput - creates annotated source and FILELIST
#
sub generateOutput
{
    my @errdsp_flags = SPLIT_STRING_ALLOWING_QUOTED_TERMS( $errdspFlags );

    if ( $integrationType =~ /STDERR/o )
    {
        my $stderrFile = APPLY_FINAL_SEP( $outputDirectory ) . BASE_NAME( $sourceFile ) . ".stderr";

        my $exitValue = EXECUTE_ARGS_WITH_RESULT
                           ( $errdspExecutable, $analyser, $sourceFile, "-hdr-",
                             "-op", $outputDirectory, @errdsp_flags, "-messagesonly+", 
                             "-file", $stderrFile );

        my @errors = LOAD_FILE_AS_ARRAY( $stderrFile );

        print STDERR @errors;
    }

    if ( $integrationType =~ /TEXT/o )
    {
        my $textFile = APPLY_FINAL_SEP( $outputDirectory ) . BASE_NAME( $sourceFile ) . ".txt";

        my $exitValue = EXECUTE_ARGS_WITH_RESULT
                           ( $errdspExecutable, $analyser, $sourceFile, "-hdr-",
                             "-op", $outputDirectory, @errdsp_flags,
                             "-file", $textFile );
    }

    if ( $integrationType =~ /HTML/o )
    {
        my $htmlFile = APPLY_FINAL_SEP( $outputDirectory ) . BASE_NAME( $sourceFile ) . ".html";

        my $exitValue = EXECUTE_ARGS_WITH_RESULT
                           ( $errdspExecutable, $analyser, $sourceFile, "-hdr-", "-html+",
                             "-op", $outputDirectory, @errdsp_flags,
                             "-file", $htmlFile );
    }

    if ( $fileList ne "" )
    {
        addToFileList();
    }
}

# ------------------------------------------------------------------------------

sub addToFileList
{
    my $full_source_filename = SHORT_PATH_NAME( GET_FILE_PATH_ABSOLUTE( $sourceFile ) );

    my $source_directory_path = SHORT_PATH_NAME( DIR_NAME( $full_source_filename ) );

    if ( ! ( -e $fileList ) )
    {
        # $fileList is already an absolute path at this point

        my $filelist_directory_path = DIR_NAME( $fileList );

        # Trim off trailing separator(s)
        #
        $filelist_directory_path =~ s/[\\\/]+$//o;

        WRITE_FILE( $fileList, ( "-op \"" . $filelist_directory_path . "\"\n",
                                 "\"" . $full_source_filename . "\"\n" )
                  );
    }
    else
    {
        # Test whether this file has already been written:

        my @file_data = LOAD_FILE_AS_ARRAY( $fileList );

        if ( grep( /^\Q\"${full_source_filename}\"\E$/, @file_data ) == 0 )
        {
            # Here when not found
            #
            push @file_data, "\"" . $full_source_filename . "\"\n";

            WRITE_FILE( $fileList, @file_data );
        }
    }
}

# ------------------------------------------------------------------------------

sub add_To_WQP_List
{
    # Usage: add_To_WQP_List( $full_source_filename );
    #
    my $full_source_filename = shift;

    if ( ! ( -e $wqp_list_file_path ) )
    {
        WRITE_FILE( $wqp_list_file_path, ( $full_source_filename . "\n" ) );
    }
    else
    {
        # Test whether this file has already been written:
        #
        my @file_data = LOAD_FILE_AS_ARRAY( $wqp_list_file_path );

        if ( grep( /^\Q${full_source_filename}\E$/, @file_data ) == 0 )
        {
            # Here when not found
            #
            push @file_data, $full_source_filename . "\n";

            WRITE_FILE( $wqp_list_file_path, @file_data );
        }
    }
}

# ------------------------------------------------------------------------------

sub Write_WQP_File
{
    # Creates a file named "$sourceFile.wqp" and writes the sourceFile
    # and options to it.
    #
    my $sourceFile = shift;
    my @options = @_;

    # First remove the -op option
    #
    my @options_less_op = ();
    my $index = 0;
    my $output_dir = PWD();

    while ( $index <= $#options )
    {
        if ( $options[ $index ] =~ /^\-(op|outputpath)$/oi )
        {
            $output_dir = $options[ $index + 1 ];

            $index += 2;
        }
        else
        {
            push @options_less_op, $options[ $index ];

            ++$index;
        }
    }

    # Now, add eol to each option.
    #
    foreach my $opt ( @options_less_op )
    {
        $opt .= "\n";
    }

    my $source_directory_path = DIR_NAME( GET_FILE_PATH_ABSOLUTE( $sourceFile ) );
    my $output_directory_path = GET_FILE_PATH_ABSOLUTE( $output_dir );

    logMessage( "Writing " . APPLY_FINAL_SEP( $source_directory_path ) . BASE_NAME( $sourceFile ) . ".wqp\n" );

    WRITE_FILE( "$sourceFile.wqp", 
                APPLY_FINAL_SEP( $source_directory_path ) . BASE_NAME( $sourceFile ) . "\n", 
                "-op\n",
                "$output_directory_path\n",
                @options_less_op );

    add_To_WQP_List( APPLY_FINAL_SEP( $source_directory_path ) . BASE_NAME( $sourceFile ) . ".wqp" );
}
 
# ------------------------------------------------------------------------------
# CR 12032. When running on cygwin, convert UNIX paths to full Windows paths
# NOTE: on cygwin running wrapper.exe the OSTYPE is MSWin32 as we are using
# ActiveState perl in a DOS subshell.
# Otherise, return (ie UNIX) just return path (ie no-op)
# Only do absolute paths - relative paths are expanded 
# to Windows paths correctly
#
sub GET_CORRECT_PATH
{
    my $path = shift;

    if( $windows_or_cygwin )
    {
        # See if path begins with a /
        #
        if ( $path =~ /^\//o )
        {
            $path = `cygpath -m "$path"`;

            $path = CHOMP( $path );

            return $path;
        }
    }

    return $path;
}

# ------------------------------------------------------------------------------
# CR 12501 Function to combine the output path with the name specified in the
# OUTPUT_DIR option.
#
sub MODIFY_OUT_PATH
{
    # Usage: $outputDirectory = MODIFY_OUT_PATH( $outputDirectory, $altOutputDir );
    #
    # $altOutputDir is either relative or absolute.
    #
    # If $altOutputDir is relative it is simply appended to the supplied
    # $outputDirectory path.
    #
    # If $altOutputDir is absolute, it is compared, element by element, with
    # $outputDirectory, starting at the lhs.  The first element that is different
    # (and any trailing elements) are appended to $altOutputDir.
    # e.g.
    # $altOutputDir is C:\A\B\QAC\M\N
    # $outputDirectory is C:\A\B\C\D
    # Then $outputDirectory becomes C:\A\B\QAC\M\N\C\D
    #
    # Similarly..
    # $altOutputDir is C:\A\B\QAC\M\N
    # $outputDirectory is C:\Fred\Blogs
    # Then $outputDirectory becomes C:\A\B\QAC\M\N\Fred\Blogs
    #
    # Note that the drive letter is treated as a directory element (if different)
    # e.g.
    # $altOutputDir is C:\A\B\QAC\M\N
    # $outputDirectory is D:\A\B\C\D
    # Then $outputDirectory becomes C:\A\B\QAC\M\N\D\A\B\C\D
    # 
    my $outputDirectory = shift;
    my $altOutputDir    = shift;

    # Test whether $altOutputDir is a relative path.
    #
    my $path_is_absolute = IS_FILE_PATH_ABSOLUTE( $altOutputDir );

    if ( ! $path_is_absolute )
    {
        # Here with relative path
        #
        return APPLY_FINAL_SEP( $outputDirectory ) . $altOutputDir;
    }

    # Here with absolute path

    # Ensure correct Cygwin path where necessary
    #
    $altOutputDir = GET_CORRECT_PATH( $altOutputDir );

    # For Cygwin and Unix platform, expand any symbolic links in the paths
    # NB wrapper.exe running on Cygwin can have symbolic links.
    #
    $outputDirectory = EXPAND_LINKS( $outputDirectory );
    $altOutputDir    = EXPAND_LINKS( $altOutputDir );

    # Decompose the paths, including the drive letter if present
    #
    my @o_path = split( /[\\\/\:]+/o, $outputDirectory );
    my @a_path = split( /[\\\/\:]+/o, $altOutputDir );

    # Force the drive letter to upper case (if present)
    #

    if ( $outputDirectory =~ /^.\:/o )
    {
        $o_path[ 0 ] = uc( $o_path[ 0 ] );
    }
   
    if ( $altOutputDir =~ /^.\:/o )
    {
        $a_path[ 0 ] = uc( $a_path[ 0 ] );
    }

    # Find first element which is different between paths
    #

    my $n_elements = scalar( @o_path );
    if ( $n_elements > scalar( @a_path ) )
    {
        $n_elements = scalar( @a_path );
    }

    my $first_different_element = 0;

    ELEMENT:
    for ( my $i_element = 0; $i_element < $n_elements; ++$i_element )
    {
        if ( $o_path[ $i_element ] ne $a_path[ $i_element ] )
        {
            $first_different_element = $i_element;

            last ELEMENT;
        }
    }

    # Now append first different element and trailing elements of
    # OutputDirectory to altOutputDir.
    #
    for ( my $i_element = $first_different_element;
          $i_element <= $#o_path;
          ++$i_element )
    {
        push @a_path, $o_path[ $i_element ];
    }

    # @a_path now has the right tokens

    # Reconstitute the path 

    my $separator = SEPARATOR();

    my $result_path = join( $separator, @a_path );

    # If $altOutputDir has a drive spec, insert the colon

    if ( $altOutputDir =~ /^.\:/o )
    {
        substr( $result_path, 1, 0 ) = ":";
    }

    return $result_path;
}

# ------------------------------------------------------------------------------

sub EXPAND_LINKS
{
    # Usage: $path_expanded = EXPAND_LINKS( $path );
    # $path is a Cygwin or Unix path
    #
    my $path = shift;

    # Get absolute version of path
    #
    $path = GET_FILE_PATH_ABSOLUTE( $path );

    # Get the drive prefix (if present)
    #

    my $drive = "";

    if ( $path =~ /^(.\:)/o )
    {
        $drive = $1;
        $path =~ s/^.\://o;
    }

    my @elements = split( /[\\\/]/o, $path );

    # Now go through element by element replacing preceding path
    # with link evaluations.
    #

    my $new_path = $drive;

    # Count from 1 because first @elements is always empty (because path is abs)
    #
    for ( my $i_el = 1; $i_el <= $#elements; ++$i_el )
    {
        my $element = $elements[ $i_el ];

        $new_path .= "/". $element;

        while ( -l $new_path )
        {
            $new_path = readlink( $new_path );

            $new_path = GET_FILE_PATH_ABSOLUTE( $new_path );
        }
    }

    return $new_path;
}

# ------------------------------------------------------------------------------
# chop line and put options and defines into hash
# Option is
#   DEFINE_ON_OPTION=fr:TEST_DEFINE=FRED mt:_REENTRANT ppc:POWERPC 
# etc
#
sub PARSE_DEFINE_OPTIONS
{
    my $line = shift;

    my %defopts;
  
    # Chop up on spaces first
    #
    my @items = SPLIT_STRING_ALLOWING_QUOTED_TERMS( $line );

    # Then put into hash if correctly formed
    #
    for my $i ( @items )
    {
        my @opt = split( /:/, $i );
        if ( @opt == 2 )
        {
            $defopts{ $opt[ 0 ] } = $opt[ 1 ];
        }    
    }

    return %defopts;
}

# ------------------------------------------------------------------------------

sub EXECUTE_ARGS_WITH_RESULT
{
    # Usage: $result_code = EXECUTE_ARGS_WITH_RESULT( @args );
    #
    my $prog_name = shift;
    my @args = @_;

    if ( $pure_windows )
    {
        my $a_parameter_has_space = 0;   # false

        foreach my $arg ( @args )
        {
            if ( index( $arg, " " ) != -1 )
            {
                $a_parameter_has_space = 1;   # true
            }

            $arg = QUOTE_ARG( $arg );
        }
  
        my $path = DIR_NAME( $prog_name );
        my $prog = BASE_NAME( $prog_name );

        if ( $path ne "" )
        {
            my $short_path = SHORT_PATH_NAME( $path );
  
            # Note that $short_path can still contain spaces
            # (see SHORT_PATH_NAME documentation), if this is the
            # case die here.
            # 
            if ( ( index( $short_path, " " ) != -1 ) &&
                  $a_parameter_has_space
               )
            {
                DIE( "Windows system limitation - " .
                     "Cannot execute \"$prog_name\" " . join( " ", @args ) );
            }
  
            $prog_name = APPLY_FINAL_SEP( $short_path )  . $prog;
        }
    }
 
    my $return_code = system( $prog_name, @args );
 
    return $return_code >> 8;
}

# ------------------------------------------------------------------------------

sub PWD
{
    # Usage: $CWD = pwd();
    #
    require Cwd;
 
    my $pwd = Cwd::getcwd();

    $pwd = GET_CORRECT_PATH( $pwd );
 
    return $pwd;
}

# ------------------------------------------------------------------------------

sub GET_FILE_PATH_ABSOLUTE
{
    # Usage: $path_adjusted = GET_FILE_PATH_ABSOLUTE( $path );
    #
    # If $path is relative, returns as absolute. 
    # If absolute returns unchanged.
    #
    my $path_param = shift;
 
    # Make sure uniform separators are used.
    #
    my $path = NORMALISE_PATH_SEPARATORS( $path_param );
 
    # Test path is legal
    #
    if ( ( $path eq "" ) ||
         ( substr( $path, 0, 1 ) eq ":" )
       )
    {
       DIE( "Illegal path \"$path_param\"." );
    }
   
    if ( IS_FILE_PATH_ABSOLUTE( $path ) )
    {
       # Have absolute path
       #
       return $path;
    }
 
    my $ch1 = substr( $path, 0, 1 );
 
    if ( $windows_or_cygwin )
    {
        my $ch2_3 = substr( $path, 1, 2 );
  
        if ( ( $ch2_3 eq ":\\" ) ||
             ( $ch2_3 eq ":/" )
           )
        {
            # Have absolute path
            #
            return $path;
        }
        else
        {
            # Here when path is relative.
            # Calc absolute path.
            #
            
            # First check case like  A:fred
            # Check that drive is the drive of the current dir.
            # If same then make relative to current dir
            # Otherwise don't know what current dir is for specified
            # drive, so fail program
            #
            
            my $current_dir_path = PWD();
            
            my $abs_path;
            
            if ( substr( $path, 1, 1 ) eq ":" )
            {
                # Here with drive spec in supplied parameter ( $path_param ).
                #
             
                # Get that drive spec.
                #
                my ( $drive_specified, $path_specified, $junk ) = split( ":", $path, 3 );
             
                # Get the currently logged drive.
                #
                my ( $current_drive, $rest ) = split( ":", $current_dir_path, 2 );
             
                # Can only resolve a relative path of a drive which is the current drive.
                # e.g. can do:
                #   current drive is C:
                #   path is          C:fred
                # Cannot do
                #   current drive is C:
                #   path is          D:fred
                # as don't know the current directory on D:
                #
             
                if ( lc( $drive_specified ) eq lc( $current_drive ) )
                {
                    $abs_path = APPLY_FINAL_SEP( $current_dir_path ) . $path_specified;
                }
                else
                {
                    DIE( "Cannot derive absolute path for \"$path_param\"." );
                }
            }
            else
            {
                # Here with no drive spec in $path, i.e. $path_param.
                #

                # Remove any leading ./ pairs
                #
                $path =~ s/^(\.\/)*//o;
     
                $abs_path = APPLY_FINAL_SEP( $current_dir_path ) . $path;
            }
           
            return $abs_path;
        }
    }
    else
    {
        # Platform is Unix ( possibly Cygwin )
        #
  
        # Here when path is relative.
        # Calc absolute path.
        #
  
        # Test for path relative to login directory,
        # i.e. starts with '~'
        #
        if ( $ch1 eq "~" )
        {
           # Here with path relative to login directory
           # Get username e.g. ~fred/dir... or ~/dir... etc
           #
           my $home;
           ( $home = $path ) =~ s/^(~.*?)\/.*/$1/o;
  
           my $remainder = substr( $path, length( $home ) );
  
           # Evaluate the $home path
           #
           my $expanded_home_path = `ls -d $home`;

           # Trim
           $expanded_home_path =~ s/^\s*(.*)\s*$/$1/;
  
           # Build the absolute path
           #
 
           # Remove final sep
           #
           $expanded_home_path =~ s/[\\\/]+$//o;

           return $expanded_home_path . $remainder;
        }
        else
        {
            my $current_dir_path = PWD();
     
            # Remove any leading ./ pairs
            #
            $path =~ s/^(\.\/)*//o;
     
            return APPLY_FINAL_SEP( $current_dir_path ) . $path;
        }
    }
}

# ------------------------------------------------------------------------------

sub SHORT_PATH_NAME
{
    # Usage: $windows_or_cygwin_short_path = SHORT_PATH_NAME( $path );
    #
    my $path = shift;
 
    if ( ! defined( $path ) ||
         ( $path eq "" )
       )
    {
        return "";
    }

    if ( $cygwin )
    {
        # Allow GetShortPathName to work on Cygwin paths as well as result is
        # ok if path starts with c drive specification (eg C:
        #
        if ( $path =~ /^.\:/o )
        {
            # Here with what looks like a DOS drive spec and path

            # Remove any trailing separator as Win32::GetShortPathName
            # objects.
            #
            $path =~ s/[\\\/]+$//o;

            require Win32;

            my $spn = Win32::GetShortPathName( $path );

            if ( defined( $spn ) &&
                 ( $spn ne "" )
               )
            {
                # Here with short path name produced a result
                # 

                # Force to Cygwin / Unix separators
                #
                $spn =~ s/\\/\//go;

                return $spn;
            }
        }

        # Here with short path name failed so return original path
        #

        # Force to Cygwin / Unix separators
        #
        $path =~ s/\\/\//go;

        return $path;
    }
    elsif ( $pure_windows )
    {
        # Here with what should be a DOS drive spec and path

        # Remove any trailing separator as Win32::GetShortPathName
        # objects.
        #
        $path =~ s/[\\\/]+$//o;

        require Win32;

        my $spn = Win32::GetShortPathName( $path );

        if ( defined( $spn ) &&
             ( $spn ne "" )
           )
        {
            # Here with short path name produced a result
            # 

            # Force to Cygwin / Unix separators
            #
            $spn =~ s/\\/\//go;

            return $spn;
        }

        # Here with short path name failed so return original path
        #

        # Force to Cygwin / Unix separators
        #
        $path =~ s/\\/\//go;

        return $path;
    }
    else
    {
        # Force to Cygwin / Unix separators
        #
        $path =~ s/\\/\//go;

        return $path;
    }
}

# ------------------------------------------------------------------------------

sub DIR_NAME
{
    # Usage: $directory_of_file_path = DIR_NAME( $path );
    #
    my $path = shift;
    
    if ( $path eq "" )
    {
       return "";
    }
 
    my $n_pos = -1;
    my $keep_value;
    my $separator = SEPARATOR();
  
    # Allow both Unix and DOS separators
    #
    $path = NORMALISE_PATH_SEPARATORS( $path );

    LOOP:
    for (;;)
    {
        $keep_value = $n_pos;
        $n_pos      = index( $path, $separator, $n_pos + 1 );
        last LOOP if ( $n_pos == -1 );
    }
    
    return substr( $path, 0, $keep_value + 1 );
}

# ------------------------------------------------------------------------------

sub IS_FILE_PATH_ABSOLUTE
{
    # Usage: $is_absolute_bool = IS_FILE_PATH_ABSOLUTE( $path );
    #
    # Returns true if path starts with \ or / or c:\
    #
    my $path = shift;
 
    if ( $path eq "" )
    {
        return 0;   # false
    }
 
    if ( $windows_or_cygwin )
    {
        # Does $path start with '\' or '/'?
        #
        my $ch_1 = substr( $path, 0, 1 );

        if ( ( $ch_1 eq "\\" ) ||
             ( $ch_1 eq "/" )
           )
        {
            # Have absolute path
            #
            return 1;   # true
        }
  
        # Does $path have drive specifier and has '\' ?
        #
        my $ch2_3 = substr( $path, 1, 2 );
  
        if ( ( $ch2_3 eq ":\\" ) ||
             ( $ch2_3 eq ":/" )
           )
        {
            # Have absolute path
            #
            return 1;    # true
        }
        else
        {
            # Here when path is relative.
            #
            return 0;  # false
        }
    }
    else
    {
        # Platform is Unix
        #
        if ( index( "~/", substr( $path, 0, 1 ) ) != -1 )
        {
            # Have absolute path
            #
            return 1;   # true
        }
        else
        {
            # Here when path is relative.
            #
            return 0;  # false
        }
    }
}

# ------------------------------------------------------------------------------

sub MK_PATH
{
    # Usage:  MK_PATH( $path_to_ensure_created );
    #
    my $path = shift;
 
    if ( $path eq "" )
    {
        return;
    }

    require File::Path;
 
    File::Path::mkpath( $path, 0, 0777 );
 
    return;
}

# ------------------------------------------------------------------------------

sub NORMALISE_PATH_SEPARATORS
{
    # Usage: $path_normalised = NORMALISE_PATH_SEPARATORS( $path );
    #
    # Swaps all occurences of \ or / for true path separator.
    #
    my $path = shift;
   
    my $separator = SEPARATOR();

    $path =~ s/\\/$separator/g;
    $path =~ s/\//$separator/g;

    return $path;
}

# ------------------------------------------------------------------------------

sub BASE_NAME
{
    # Usage: $stem_and_extension = BASE_NAME( $path );
    #
    my $path = shift;

    require File::Basename;

    return File::Basename::basename( $path );
}

# ------------------------------------------------------------------------------

sub STEM_NAME
{
   # Usage: $stem = STEM_NAME( $path );
   #
   my $base_name = BASE_NAME( shift );

   my $dot_pos = rindex( $base_name, "." );

   if ( $dot_pos == -1 )
   {
      return $base_name;
   }
   elsif ( $dot_pos == 0 )
   {
      return "";
   }
   else
   {
      return substr( $base_name, 0, $dot_pos );
   }
}

# ------------------------------------------------------------------------------

sub QUOTE_ARG
{
    # Usage: $quoted_arg = QUOTE_ARG( $arg );
    #
    # Escapes any embedded quotes.
    # If $arg has embedded spaces or embedded \ , and $arg not already quoted,
    # quotes $arg.
    #
    my $arg_supplied = shift;
 
    if ( $arg_supplied eq "" )
    {
        return "";
    }
 
    # If $arg already quoted, remove quotes.
    #
    my $arg = UNQUOTE( $arg_supplied );
 
    my $need_quotes = $arg ne $arg_supplied;
    
    # Test for embedded \ in string that is NOT quoted.
    #
    if ( ( $arg eq $arg_supplied ) &&
         ( index( $arg_supplied, "\\" ) != -1 )
       )
    {
        # Here with arg_supplied is not quoted and has embedded \
        #
        # Set $need_quotes to true to force generation of surrounding quotes
        #
        $need_quotes = 1;  # true
    }
    
    # Here with $arg as bare string that may need quotes if it has
    # embedded spaces.
    #
 
    # Need to maintain any \\ or \" already in string.
    # Hence separate $arg on \\ and \"
    #
 
    my @args;
    if ( $arg eq "" )
    {
        @args = ( "" );
    }
    else
    {
        @args = split( (/(\\\\|\\\")/o), $arg );
    }
 
    # Cannot allow $arg to end in ...\ as this will escape the 
    # applied quote and corrupt the command-line.
    # Similarly settings.via file does not allow paths that end in a
    # bare \
    # e.g. can't have -opC:\prqa\qac-5.0\temp\
    # settings.via can have -op"C:\prqa\qac-5.0\temp\" but this is not
    # suitable as an argument to a Perl program.
    # Achieve this by removing last term of @args if it is '\'.
    #
 
    my $last_arg = $args[ $#args ];
 
    if ( ( $last_arg ne "\\\\" ) &&
         ( $last_arg =~ /\\$/o )
       )
    {
        ( $args[ $#args ] = $last_arg ) =~ s/\\$//o;
    }
 
    SUB_ARG:
    foreach my $sub_arg ( @args )
    {
        next SUB_ARG if ( ( $sub_arg eq "\\\\" ) ||
                          ( $sub_arg eq "\\\"" )
                        );
 
        # Escape all embedded quotes
        #
        $sub_arg =~ s/\"/\\\"/go;
    }
 
    # Reinstate $arg by reassembling the components
    #
    $arg = join( "", @args );
    
    # Now quote if necessary.
    #
    if ( $need_quotes ||
         ( $arg =~ /[\s\"]/oi ) ||
         # Add quotes if $arg contains ( or )
         ( $arg =~ /[\(\)]/o )
       )
    {
        # If $arg starts with -, put the first quote after it
        #
        if ( $arg =~ /^\-/o )
        {
            $arg =~ s/^\-([a-zA-Z0-9_=]+)(.*)$/\-$1"$2"/o;

            return $arg;
        }
        else
        {
            return "\"$arg\"";
        }
    }
    else
    {   
        # Here when no quoting necessary
        #
        return $arg;
    }
}

# ------------------------------------------------------------------------------

sub UNQUOTE
{
    # Usage: $unquoted_string = UNQUOTE( $arg );
    #
    my $arg = shift;
 
    if ( IS_QUOTED( $arg, "\"" ) )
    {
        return substr( $arg, 1, length( $arg ) - 2 );
    }
    else
    {
        return $arg;
    }
}

# ------------------------------------------------------------------------------

sub IS_QUOTED
{
    # Usage: $bool = IS_QUOTED( $item, "\'\"" );
    #
    my $item   = shift;
    my $quotes = shift;
    
    if ( length( $item ) < 2 )
    {
        return 0;  # False
    }
 
    my $first_ch = substr( $item, 0, 1 );
    my $last_ch  = substr( $item, length( $item ) - 1, 1 );
 
    my @quote_chars = split( //, $quotes );
 
    foreach my $ch ( @quote_chars )
    {
        if ( ( $first_ch eq $ch ) && ( $last_ch eq $ch ) )
        {
            return 1;  # true
        }
    }
 
    return 0;   # false
}

# ------------------------------------------------------------------------------

sub SPLIT_STRING_ALLOWING_QUOTED_TERMS
{
    # Subroutine to split up a string of terms separated by space characters.
    # Where some of the terms may be quoted strings that include spaces.
    # Note that escaped spaces may be embedded.  To do this the space is 
    # prefixed with \ (backslash).
    #
    # Usage:
    #  @list_of_terms = SPLIT_STRING_ALLOWING_QUOTED_TERMS( $str );
    #
    my $str = shift;

    if ( $str eq "" )
    {
        return ();
    }
 
    my $separator_char = " ";                             # Constant
    my $escape_char    = "\\";                            # Constant
    my $double_quote   = "\"";                            # Constant
    my $single_quote   = "'";                             # Constant
    my $quoting_chars  = $double_quote . $single_quote;   # Constant
    my $escaped_chars  = $double_quote . $escape_char;    # Constant
    
 
    my $in_quoted_term = 0;   # FALSE
    my $term           = "";
    my @list_of_terms  = ();
    my @characters     = split( //, $str );
 
    my $quote_char = "";
    my $i_ch = 0;
 
    I_CH:
    while ( $i_ch <= $#characters )
    {
        my $ch = $characters[ $i_ch ];
 
        my $next_ch;
        if ( $i_ch < $#characters )
        {
            $next_ch = $characters[ $i_ch + 1 ];
        }
        else
        {
            $next_ch = "";
        }
 
        if ( index( $quoting_chars, $ch ) != -1 )
        {
            if ( $in_quoted_term )
            {
                if ( $ch eq $quote_char )
                {
                    $in_quoted_term = 0;  # False
                }
                else
                {
                    $term .= $ch;
                }
            }
            else
            {
                $in_quoted_term = 1;   # True
                $quote_char = $ch;
            }
        }
        elsif ( $ch eq $escape_char )
        {
            # Here with escape chars.
            # "abc\"def\\ghi\ jk\'l\x\y\z" becomes -> abc"def\ghi\ jk\'l\x\y\z
            # 'abc\"def\\ghi\ jk\'l\x\y\z' is illegal (unterminated)
            # 'abc\"def\\ghi\ jkl\x\y\z'   becomes -> abc\"def\\ghi\ jkl\x\y\z
            # abc\"def\\ghi\ jk\'l\x\y\z   becomes -> abc"def\ghi jk'lxyz
            #
            
            # Therefore, escape char consumed if outside of quotes or
            # if inside double quotes and escaped char is " or \
            #
            if ( ! $in_quoted_term ||
                 ( $in_quoted_term &&
                   ( $quote_char eq $double_quote ) &&
                   ( index( $escaped_chars, $next_ch ) != -1 )
                 )
               )
            {
                # Drop the escape char and use the next char
                #
                $term .= $next_ch;
                ++$i_ch;
            }
            else
            {
                # Treat as ordinary char
                #
                $term .= $ch;
            }
        }
        elsif ( $ch eq $separator_char )
        {
            if ( $in_quoted_term )
            {
                $term .= $ch;
            }
            else
            {
                if ( $term ne "" )
                {
                    push @list_of_terms, $term;
                    $term = "";
                }
            }
        }
        else
        {
            $term .= $ch;
        }
 
        ++$i_ch;
    }
 
    if ( $term ne "" )
    {
        push @list_of_terms, $term;
    }
 
    return @list_of_terms;
}

# ------------------------------------------------------------------------------

sub APPLY_FINAL_SEP
{
    # e.g. $qabin = APPLY_FINAL_SEP( $QACBIN );
    #
    my $path = shift;

    my $path_separator = SEPARATOR();

    if ( $path eq "" )
    {
        return $path;
    }

    # Trim any existing trailing separators
    #
    $path =~ s/[\/\\]*$//o;

    # Append separator
    #
    $path .= $path_separator;

    # Force separators to '/'
    #
    $path = NORMALISE_PATH_SEPARATORS( $path );
    
    return $path;
}

# ------------------------------------------------------------------------------

sub SEPARATOR
{
    # Usage: $separator = SEPARATOR();
    #

    # Always return '/' as this works for Windows QAC as well as Cygwin and Unix.
    #
    return "/";

    if ( $pure_windows )
    {
        return "\\";
    }
    else
    {
        return "/";
    }
}

# ------------------------------------------------------------------------------

sub CHOMP
{
   # Usage: $line = CHOMP( $line );
   # chomp implemented such that it will work for Unix and Windows eol
   #
   my $line = shift;

   $line =~ s/[\n\r]+$//o;

   return $line;
}

# ------------------------------------------------------------------------------

sub LOAD_FILE_AS_ARRAY
{
   # Usage @array_of_lines_of_file = LOAD_FILE_AS_ARRAY( $file_name );
   #
   my $file_name = shift;

   open FILE, "<$file_name" or DIE( "Cannot open file $file_name\n$!" );
   
   my @file_lines = <FILE>;
 
   close( FILE );
   
   return @file_lines;
}

# ------------------------------------------------------------------------------

sub WRITE_FILE
{
   # Usage: WRITE_FILE( $file_name, @text );
   #
   my $file_name = shift;
   my @text      = @_;

   open FILE, ">$file_name" or DIE( "Cannot open $file_name\n$!" );

   foreach my $line ( @text )
   {
       print FILE $line;
   }

   close FILE;
}

# ------------------------------------------------------------------------------

sub DIE
{
    # Usage: DIE( $str );
    #
    my $str = shift;

    print STDERR ">> Error - $str\n";

    exit 255;
}

# ------------------------------------------------------------------------------

sub WINDOWS_ONLY_WHICH_PROGRAM
{
    # Usage: $full_filename = WINDOWS_ONLY_WHICH_PROGRAM( $program_name );
    #
    # Returns absolute path to $program_name or "" if not found.
    #
    # NB this WHICH() is identical to GEN::WHICH() and has the limitation that
    # for Windows, if the supplied file is not an executable file, e.g. fred.txt,
    # then WHICH() will return the first full path of fred.txt in
    # PATH, whether it is executable or not.  It is the responsibility
    # of the user of WHICH() to ensure that a truely executable
    # filename is supplied.
    #
    # For Windows:
    #    Gets all paths on PATH and for each path:
    #       Looks for $program_name . ( '.exe' | '.com' | '' ).
    #
    # NB For Windows, this sub will return the first file path in PATH
    # that matches $program_name . ( '.exe' | '.com' | '' )
    # This may be a non-executable file.
  
    my $program_name = shift;
 
    if ( ! $windows_or_cygwin )
    {
        # This sub cannot be called for Unix

        DIE( "Program error in WINDOWS_ONLY_WHICH_PROGRAM" );
    }
    
    if ( $program_name eq "" )
    {
        return "";
    }
 
    # No need to lookup if $program_name is absolute.
    #
    elsif ( IS_FILE_PATH_ABSOLUTE( $program_name ) )
    {
        return EXECUTABLE_FULL_NAME( $program_name );
    }

    # Here when on Windows.
    #
    else
    {
        my $prog       = BASE_NAME( $program_name );
        my $prog_stem  = STEM_NAME( $program_name );
  
        my @exe_names = ( "$prog.exe", "$prog.com", $prog );
  
        my $path_delimiter = ";";
        if ( $cygwin )
        {
            $path_delimiter = ":";
        }

        my $path_list = $ENV{ 'PATH' };
        $path_list    = "." . $path_delimiter . $path_list;
  
        my @paths = split( $path_delimiter, $path_list );
    
        foreach my $path ( @paths )
        {
            $path = APPLY_FINAL_SEP( $path );

            foreach my $exe_name ( @exe_names )
            {
                my $try_name = $path . $exe_name;
    
                if ( -f $try_name )
                {
                    return EXECUTABLE_FULL_NAME( $try_name );
                }
            }
        }
    }
 
    return "";
}

# ------------------------------------------------------------------------------

sub EXECUTABLE_FULL_NAME
{
   # Usage: $full_program_name = EXECUTABLE_FULL_NAME( $program_name );
   #
   my $program_name = shift;

   my $executable_full_name = "";

   if ( $unix )
   {
      $executable_full_name = $program_name;
   }
   else
   {
      if ( -e "$program_name.exe" )
      {
         $executable_full_name = "$program_name.exe";
      }
      if ( -e "$program_name.com" )
      {
         $executable_full_name = "$program_name.com";
      }
      elsif ( -e $program_name )
      {
         $executable_full_name = $program_name;
      }
   }

   if ( $executable_full_name eq "" )
   {
      return "";
   }

   ### $executable_full_name = GET_FILE_PATH_ABSOLUTE( $executable_full_name );
   ### $executable_full_name = EVALUATE_RELATIVE_PATHS( $executable_full_name );

   return $executable_full_name;
}

# ------------------------------------------------------------------------------
